/* This file is autogenerated by running:
 * scripts/keymap-render.py > src/libply-splash-graphics/ply-keymap-metadata.h
 */
typedef struct
{
        const char *name;
        int         offset;
        int         width;
        int         type;
} ply_keymap_metadata_t;

typedef enum
{
        PLY_LAYOUT_TERMINAL,
        PLY_LAYOUT_XKB,
} ply_layout_types_t;

static ply_keymap_metadata_t ply_keymap_metadata[] = {
        { "3l",                                        10,    47,  PLY_LAYOUT_TERMINAL },
        { "dvorak",                                    57,    113, PLY_LAYOUT_TERMINAL },
        { "adnw",                                      170,   94,  PLY_LAYOUT_TERMINAL },
        { "al",                                        264,   45,  PLY_LAYOUT_TERMINAL },
        { "amiga",                                     309,   103, PLY_LAYOUT_TERMINAL },
        { "apple",                                     412,   93,  PLY_LAYOUT_TERMINAL },
        { "applkey",                                   505,   126, PLY_LAYOUT_TERMINAL },
        { "ara",                                       631,   62,  PLY_LAYOUT_TERMINAL },
        { "at",                                        693,   46,  PLY_LAYOUT_TERMINAL },
        { "atari",                                     739,   81,  PLY_LAYOUT_TERMINAL },
        { "az",                                        820,   49,  PLY_LAYOUT_TERMINAL },
        { "azerty",                                    869,   105, PLY_LAYOUT_TERMINAL },
        { "ba",                                        974,   50,  PLY_LAYOUT_TERMINAL },
        { "backspace",                                 1024,  157, PLY_LAYOUT_TERMINAL },
        { "bashkir",                                   1181,  119, PLY_LAYOUT_TERMINAL },
        { "be",                                        1300,  50,  PLY_LAYOUT_TERMINAL },
        { "bg",                                        1350,  51,  PLY_LAYOUT_TERMINAL },
        { "bone",                                      1401,  84,  PLY_LAYOUT_TERMINAL },
        { "br",                                        1485,  48,  PLY_LAYOUT_TERMINAL },
        { "by",                                        1533,  51,  PLY_LAYOUT_TERMINAL },
        { "bywin",                                     1584,  98,  PLY_LAYOUT_TERMINAL },
        { "ca",                                        1682,  48,  PLY_LAYOUT_TERMINAL },
        { "carpalx",                                   1730,  118, PLY_LAYOUT_TERMINAL },
        { "cf",                                        1848,  46,  PLY_LAYOUT_TERMINAL },
        { "ch",                                        1894,  49,  PLY_LAYOUT_TERMINAL },
        { "cm",                                        1943,  59,  PLY_LAYOUT_TERMINAL },
        { "cn",                                        2002,  49,  PLY_LAYOUT_TERMINAL },
        { "croat",                                     2051,  89,  PLY_LAYOUT_TERMINAL },
        { "ctrl",                                      2140,  66,  PLY_LAYOUT_TERMINAL },
        { "cz",                                        2206,  47,  PLY_LAYOUT_TERMINAL },
        { "de",                                        2253,  51,  PLY_LAYOUT_TERMINAL },
        { "defkeymap",                                 2304,  170, PLY_LAYOUT_TERMINAL },
        { "dk",                                        2474,  53,  PLY_LAYOUT_TERMINAL },
        { "dz",                                        2527,  50,  PLY_LAYOUT_TERMINAL },
        { "ee",                                        2577,  50,  PLY_LAYOUT_TERMINAL },
        { "emacs",                                     2627,  106, PLY_LAYOUT_TERMINAL },
        { "emacs2",                                    2733,  122, PLY_LAYOUT_TERMINAL },
        { "en",                                        2855,  51,  PLY_LAYOUT_TERMINAL },
        { "epo",                                       2906,  69,  PLY_LAYOUT_TERMINAL },
        { "es",                                        2975,  49,  PLY_LAYOUT_TERMINAL },
        { "et",                                        3024,  46,  PLY_LAYOUT_TERMINAL },
        { "euro",                                      3070,  81,  PLY_LAYOUT_TERMINAL },
        { "euro1",                                     3151,  92,  PLY_LAYOUT_TERMINAL },
        { "euro2",                                     3243,  97,  PLY_LAYOUT_TERMINAL },
        { "fa",                                        3340,  46,  PLY_LAYOUT_TERMINAL },
        { "fi",                                        3386,  38,  PLY_LAYOUT_TERMINAL },
        { "fo",                                        3424,  48,  PLY_LAYOUT_TERMINAL },
        { "fr",                                        3472,  44,  PLY_LAYOUT_TERMINAL },
        { "gb",                                        3516,  54,  PLY_LAYOUT_TERMINAL },
        { "ge",                                        3570,  52,  PLY_LAYOUT_TERMINAL },
        { "gh",                                        3622,  53,  PLY_LAYOUT_TERMINAL },
        { "gr",                                        3675,  50,  PLY_LAYOUT_TERMINAL },
        { "hr",                                        3725,  48,  PLY_LAYOUT_TERMINAL },
        { "hu",                                        3773,  50,  PLY_LAYOUT_TERMINAL },
        { "hu101",                                     3823,  96,  PLY_LAYOUT_TERMINAL },
        { "id",                                        3919,  43,  PLY_LAYOUT_TERMINAL },
        { "ie",                                        3962,  42,  PLY_LAYOUT_TERMINAL },
        { "il",                                        4004,  37,  PLY_LAYOUT_TERMINAL },
        { "in",                                        4041,  43,  PLY_LAYOUT_TERMINAL },
        { "iq",                                        4084,  43,  PLY_LAYOUT_TERMINAL },
        { "ir",                                        4127,  40,  PLY_LAYOUT_TERMINAL },
        { "is",                                        4167,  41,  PLY_LAYOUT_TERMINAL },
        { "it",                                        4208,  38,  PLY_LAYOUT_TERMINAL },
        { "it2",                                       4246,  54,  PLY_LAYOUT_TERMINAL },
        { "jp",                                        4300,  46,  PLY_LAYOUT_TERMINAL },
        { "jp106",                                     4346,  96,  PLY_LAYOUT_TERMINAL },
        { "kazakh",                                    4442,  113, PLY_LAYOUT_TERMINAL },
        { "ke",                                        4555,  49,  PLY_LAYOUT_TERMINAL },
        { "keypad",                                    4604,  114, PLY_LAYOUT_TERMINAL },
        { "ko",                                        4718,  51,  PLY_LAYOUT_TERMINAL },
        { "koy",                                       4769,  67,  PLY_LAYOUT_TERMINAL },
        { "kr",                                        4836,  47,  PLY_LAYOUT_TERMINAL },
        { "ky",                                        4883,  50,  PLY_LAYOUT_TERMINAL },
        { "kyrgyz",                                    4933,  108, PLY_LAYOUT_TERMINAL },
        { "kz",                                        5041,  48,  PLY_LAYOUT_TERMINAL },
        { "la",                                        5089,  42,  PLY_LAYOUT_TERMINAL },
        { "latam",                                     5131,  96,  PLY_LAYOUT_TERMINAL },
        { "lk",                                        5227,  44,  PLY_LAYOUT_TERMINAL },
        { "lt",                                        5271,  38,  PLY_LAYOUT_TERMINAL },
        { "lv",                                        5309,  43,  PLY_LAYOUT_TERMINAL },
        { "ma",                                        5352,  60,  PLY_LAYOUT_TERMINAL },
        { "no",                                        5412,  52,  PLY_LAYOUT_TERMINAL },
        { "pl",                                        5464,  45,  PLY_LAYOUT_TERMINAL },
        { "pt",                                        5509,  46,  PLY_LAYOUT_TERMINAL },
        { "se",                                        5555,  48,  PLY_LAYOUT_TERMINAL },
        { "template",                                  5603,  142, PLY_LAYOUT_TERMINAL },
        { "uk",                                        5745,  53,  PLY_LAYOUT_TERMINAL },
        { "us",                                        5798,  50,  PLY_LAYOUT_TERMINAL },
        { "md",                                        5848,  61,  PLY_LAYOUT_TERMINAL },
        { "me",                                        5909,  60,  PLY_LAYOUT_TERMINAL },
        { "mk",                                        5969,  62,  PLY_LAYOUT_TERMINAL },
        { "mk0",                                       6031,  79,  PLY_LAYOUT_TERMINAL },
        { "ml",                                        6110,  55,  PLY_LAYOUT_TERMINAL },
        { "mm",                                        6165,  71,  PLY_LAYOUT_TERMINAL },
        { "mt",                                        6236,  56,  PLY_LAYOUT_TERMINAL },
        { "neo",                                       6292,  68,  PLY_LAYOUT_TERMINAL },
        { "neoqwertz",                                 6360,  162, PLY_LAYOUT_TERMINAL },
        { "ng",                                        6522,  51,  PLY_LAYOUT_TERMINAL },
        { "nl",                                        6573,  45,  PLY_LAYOUT_TERMINAL },
        { "nl2",                                       6618,  60,  PLY_LAYOUT_TERMINAL },
        { "pc110",                                     6678,  95,  PLY_LAYOUT_TERMINAL },
        { "ph",                                        6773,  51,  PLY_LAYOUT_TERMINAL },
        { "pl1",                                       6824,  55,  PLY_LAYOUT_TERMINAL },
        { "pl2",                                       6879,  60,  PLY_LAYOUT_TERMINAL },
        { "pl3",                                       6939,  61,  PLY_LAYOUT_TERMINAL },
        { "pl4",                                       7000,  62,  PLY_LAYOUT_TERMINAL },
        { "ro",                                        7062,  47,  PLY_LAYOUT_TERMINAL },
        { "rs",                                        7109,  44,  PLY_LAYOUT_TERMINAL },
        { "ru",                                        7153,  45,  PLY_LAYOUT_TERMINAL },
        { "ru1",                                       7198,  58,  PLY_LAYOUT_TERMINAL },
        { "ru2",                                       7256,  63,  PLY_LAYOUT_TERMINAL },
        { "ru3",                                       7319,  64,  PLY_LAYOUT_TERMINAL },
        { "ru4",                                       7383,  65,  PLY_LAYOUT_TERMINAL },
        { "ruwin",                                     7448,  95,  PLY_LAYOUT_TERMINAL },
        { "sg",                                        7543,  49,  PLY_LAYOUT_TERMINAL },
        { "si",                                        7592,  40,  PLY_LAYOUT_TERMINAL },
        { "sk",                                        7632,  50,  PLY_LAYOUT_TERMINAL },
        { "slovene",                                   7682,  122, PLY_LAYOUT_TERMINAL },
        { "sr",                                        7804,  46,  PLY_LAYOUT_TERMINAL },
        { "sunkeymap",                                 7850,  174, PLY_LAYOUT_TERMINAL },
        { "sv",                                        8024,  49,  PLY_LAYOUT_TERMINAL },
        { "sy",                                        8073,  49,  PLY_LAYOUT_TERMINAL },
        { "tj",                                        8122,  38,  PLY_LAYOUT_TERMINAL },
        { "tm",                                        8160,  57,  PLY_LAYOUT_TERMINAL },
        { "tr",                                        8217,  44,  PLY_LAYOUT_TERMINAL },
        { "tralt",                                     8261,  79,  PLY_LAYOUT_TERMINAL },
        { "trf",                                       8340,  56,  PLY_LAYOUT_TERMINAL },
        { "trq",                                       8396,  59,  PLY_LAYOUT_TERMINAL },
        { "ttwin",                                     8455,  90,  PLY_LAYOUT_TERMINAL },
        { "tw",                                        8545,  56,  PLY_LAYOUT_TERMINAL },
        { "ua",                                        8601,  51,  PLY_LAYOUT_TERMINAL },
        { "unicode",                                   8652,  124, PLY_LAYOUT_TERMINAL },
        { "us1",                                       8776,  61,  PLY_LAYOUT_TERMINAL },
        { "uz",                                        8837,  50,  PLY_LAYOUT_TERMINAL },
        { "vn",                                        8887,  51,  PLY_LAYOUT_TERMINAL },
        { "wangbe",                                    8938,  127, PLY_LAYOUT_TERMINAL },
        { "wangbe2",                                   9065,  144, PLY_LAYOUT_TERMINAL },
        { "windowkeys",                                9209,  188, PLY_LAYOUT_TERMINAL },
        { "English (US)",                              9397,  182, PLY_LAYOUT_XKB      },
        { "Dari",                                      9579,  75,  PLY_LAYOUT_XKB      },
        { "Arabic",                                    9654,  107, PLY_LAYOUT_XKB      },
        { "Albanian",                                  9761,  139, PLY_LAYOUT_XKB      },
        { "Armenian",                                  9900,  152, PLY_LAYOUT_XKB      },
        { "German (Austria)",                          10052, 252, PLY_LAYOUT_XKB      },
        { "English (Australian)",                      10304, 284, PLY_LAYOUT_XKB      },
        { "Azerbaijani",                               10588, 172, PLY_LAYOUT_XKB      },
        { "Belarusian",                                10760, 162, PLY_LAYOUT_XKB      },
        { "Belgian",                                   10922, 121, PLY_LAYOUT_XKB      },
        { "Bangla",                                    11043, 113, PLY_LAYOUT_XKB      },
        { "Indian",                                    11156, 101, PLY_LAYOUT_XKB      },
        { "Bosnian",                                   11257, 126, PLY_LAYOUT_XKB      },
        { "Portuguese (Brazil)",                       11383, 279, PLY_LAYOUT_XKB      },
        { "Bulgarian",                                 11662, 150, PLY_LAYOUT_XKB      },
        { "Berber (Algeria, Latin)",                   11812, 317, PLY_LAYOUT_XKB      },
        { "Arabic (Morocco)",                          12129, 249, PLY_LAYOUT_XKB      },
        { "English (Cameroon)",                        12378, 285, PLY_LAYOUT_XKB      },
        { "Burmese",                                   12663, 139, PLY_LAYOUT_XKB      },
        { "French (Canada)",                           12802, 237, PLY_LAYOUT_XKB      },
        { "French (Democratic Republic of the Congo)", 13039, 594, PLY_LAYOUT_XKB      },
        { "Chinese",                                   13633, 125, PLY_LAYOUT_XKB      },
        { "Croatian",                                  13758, 134, PLY_LAYOUT_XKB      },
        { "Czech",                                     13892, 99,  PLY_LAYOUT_XKB      },
        { "Danish",                                    13991, 111, PLY_LAYOUT_XKB      },
        { "Dutch",                                     14102, 98,  PLY_LAYOUT_XKB      },
        { "Dzongkha",                                  14200, 155, PLY_LAYOUT_XKB      },
        { "Estonian",                                  14355, 135, PLY_LAYOUT_XKB      },
        { "Persian",                                   14490, 118, PLY_LAYOUT_XKB      },
        { "Iraqi",                                     14608, 79,  PLY_LAYOUT_XKB      },
        { "Faroese",                                   14687, 125, PLY_LAYOUT_XKB      },
        { "Finnish",                                   14812, 115, PLY_LAYOUT_XKB      },
        { "French",                                    14927, 110, PLY_LAYOUT_XKB      },
        { "English (Ghana)",                           15037, 232, PLY_LAYOUT_XKB      },
        { "N'Ko (AZERTY)",                             15269, 225, PLY_LAYOUT_XKB      },
        { "Georgian",                                  15494, 144, PLY_LAYOUT_XKB      },
        { "German",                                    15638, 128, PLY_LAYOUT_XKB      },
        { "Greek",                                     15766, 102, PLY_LAYOUT_XKB      },
        { "Hungarian",                                 15868, 160, PLY_LAYOUT_XKB      },
        { "Icelandic",                                 16028, 138, PLY_LAYOUT_XKB      },
        { "Hebrew",                                    16166, 126, PLY_LAYOUT_XKB      },
        { "Italian",                                   16292, 103, PLY_LAYOUT_XKB      },
        { "Japanese",                                  16395, 144, PLY_LAYOUT_XKB      },
        { "Kyrgyz",                                    16539, 112, PLY_LAYOUT_XKB      },
        { "Khmer (Cambodia)",                          16651, 272, PLY_LAYOUT_XKB      },
        { "Kazakh",                                    16923, 117, PLY_LAYOUT_XKB      },
        { "Lao",                                       17040, 67,  PLY_LAYOUT_XKB      },
        { "Spanish (Latin American)",                  17107, 355, PLY_LAYOUT_XKB      },
        { "Lithuanian",                                17462, 160, PLY_LAYOUT_XKB      },
        { "Latvian",                                   17622, 116, PLY_LAYOUT_XKB      },
        { "Maori",                                     17738, 96,  PLY_LAYOUT_XKB      },
        { "Montenegrin",                               17834, 192, PLY_LAYOUT_XKB      },
        { "Macedonian",                                18026, 181, PLY_LAYOUT_XKB      },
        { "Maltese",                                   18207, 125, PLY_LAYOUT_XKB      },
        { "Mongolian",                                 18332, 162, PLY_LAYOUT_XKB      },
        { "Norwegian",                                 18494, 168, PLY_LAYOUT_XKB      },
        { "Polish",                                    18662, 100, PLY_LAYOUT_XKB      },
        { "Portuguese",                                18762, 173, PLY_LAYOUT_XKB      },
        { "Romanian",                                  18935, 154, PLY_LAYOUT_XKB      },
        { "Russian",                                   19089, 122, PLY_LAYOUT_XKB      },
        { "Serbian",                                   19211, 122, PLY_LAYOUT_XKB      },
        { "Slovenian",                                 19333, 151, PLY_LAYOUT_XKB      },
        { "Slovak",                                    19484, 111, PLY_LAYOUT_XKB      },
        { "Spanish",                                   19595, 125, PLY_LAYOUT_XKB      },
        { "Swedish",                                   19720, 132, PLY_LAYOUT_XKB      },
        { "German (Switzerland)",                      19852, 316, PLY_LAYOUT_XKB      },
        { "Arabic (Syria)",                            20168, 200, PLY_LAYOUT_XKB      },
        { "Tajik",                                     20368, 87,  PLY_LAYOUT_XKB      },
        { "Sinhala (phonetic)",                        20455, 262, PLY_LAYOUT_XKB      },
        { "Thai",                                      20717, 78,  PLY_LAYOUT_XKB      },
        { "Turkish",                                   20795, 121, PLY_LAYOUT_XKB      },
        { "Taiwanese",                                 20916, 164, PLY_LAYOUT_XKB      },
        { "Ukrainian",                                 21080, 149, PLY_LAYOUT_XKB      },
        { "English (UK)",                              21229, 185, PLY_LAYOUT_XKB      },
        { "Uzbek",                                     21414, 105, PLY_LAYOUT_XKB      },
        { "Vietnamese",                                21519, 179, PLY_LAYOUT_XKB      },
        { "Korean",                                    21698, 115, PLY_LAYOUT_XKB      },
        { "Irish",                                     21813, 77,  PLY_LAYOUT_XKB      },
        { "Urdu (Pakistan)",                           21890, 227, PLY_LAYOUT_XKB      },
        { "Dhivehi",                                   22117, 120, PLY_LAYOUT_XKB      },
        { "English (South Africa)",                    22237, 310, PLY_LAYOUT_XKB      },
        { "Esperanto",                                 22547, 156, PLY_LAYOUT_XKB      },
        { "Nepali",                                    22703, 106, PLY_LAYOUT_XKB      },
        { "English (Nigeria)",                         22809, 245, PLY_LAYOUT_XKB      },
        { "Amharic",                                   23054, 134, PLY_LAYOUT_XKB      },
        { "Wolof",                                     23188, 106, PLY_LAYOUT_XKB      },
        { "Braille",                                   23294, 107, PLY_LAYOUT_XKB      },
        { "Turkmen",                                   23401, 142, PLY_LAYOUT_XKB      },
        { "Bambara",                                   23543, 141, PLY_LAYOUT_XKB      },
        { "Swahili (Tanzania)",                        23684, 267, PLY_LAYOUT_XKB      },
        { "French (Togo)",                             23951, 206, PLY_LAYOUT_XKB      },
        { "Swahili (Kenya)",                           24157, 228, PLY_LAYOUT_XKB      },
        { "Tswana",                                    24385, 124, PLY_LAYOUT_XKB      },
        { "Filipino",                                  24509, 119, PLY_LAYOUT_XKB      },
        { "Moldavian",                                 24628, 158, PLY_LAYOUT_XKB      },
        { "Indonesian (Latin)",                        24786, 259, PLY_LAYOUT_XKB      },
        { "Malay (Jawi, Arabic Keyboard)",             25045, 423, PLY_LAYOUT_XKB      },
        { "A user-defined custom Layout",              25468, 423, PLY_LAYOUT_XKB      },
        { NULL,                                        } /* End of array marker */
};
